/* -*-c-mode-*- */
/*------------------------------------------------------
 *  PERF_METER.C - Performance Meter display for DECWindows
 *  Robert Heller Created on Wed Nov  8 10:39:17 1989
 *  Last mod -
 *      Mark Pizzolato 11/22/89 - mark@infopiz.uucp (uunet!lupine!infopiz!mark)
 *	Robert Heller Mon Nov 27 12:27:53 1989 - fixed for invert displays
 *------------------------------------------------------*/

#include <stdio.h>                              /* For printf and so on. */
#include <ctype.h>
#include <descrip.h>
#include <ssdef.h>
#include <jpidef.h>
#include <dvidef.h>
#include <syidef.h>
#include "cpudef.h"

globalref long int EXE$GL_ABSTIM_TICS;
globalref long int SMP$GL_ACTIVE_CPUS;
globalref $CPU *SMP$GL_CPU_DATA[1];

globalref long int PMS$GL_FAULTS;
globalref long int PMS$GL_PREADS;
globalref long int PMS$GL_PREADIO;
globalref long int PMS$GL_PWRITES;
globalref long int PMS$GL_PWRITIO;
globalref long int PMS$GL_DZROFLTS;


globalref long int SCH$GL_FREECNT;
globalref long int SCH$GL_MFYCNT;

#define current_faults (PMS$GL_FAULTS)
#define current_pageio (PMS$GL_PREADIO + PMS$GL_PWRITIO)
#define current_freepages (SCH$GL_FREECNT)
#define current_modpages (SCH$GL_MFYCNT)


#ifdef VMS
#include <decw$include/DwtAppl.h>       /* DECwindows Toolkit */
#else
#include <X11/DwtAppl.h>                /* DECwindows Toolkit */
#endif

#define FAILED(status) (((status)&1) == 0)
#define SIGNAL_ERROR(status) 				\
    {							\
    int ___status = status;				\
							\
    if (FAILED(___status))				\
	lib$signal(___status);				\
    }

#include "perf_meter_const.h"


#define MAX_WIDGETS (MAX_WIDGET + 1)

/*
 * Global data
 */

static Widget toplevel_widget,          /* Root widget ID of our */
                                        /* application. */
  main_window_widget,                   /* Root widget ID of main */
                                        /* DRM fetch */
  generic_message_widget,		/* generic message box */
  generic_caution_widget,		/* generic caution box */
  generic_wip_widget;			/* generic WIP box */
noshare widget_array[MAX_WIDGETS];            /* Place to keep all other */
                                        /* widget IDs */

static Display *cpu_graph_disp, *state_graph_disp;
static Window   cpu_graph_wind = 0,  state_graph_wind = 0;
static GC       cpu_graph_gc1,    state_graph_gc1;
static GC       cpu_graph_gc0,    state_graph_gc0;


static DwtCompString latin_create;      /* Variables for */
static DwtCompString latin_dismiss;     /* compound strings. */
static DwtCompString latin_space;
static DwtCompString latin_zero;
static DRMHierarchy s_DRMHierarchy;     /* DRM database hierarchy ID */
static DRMType *dummy_class;            /* and class variable. */
static char *db_filename_vec[] =        /* DRM heirachy file list. */
  {"pfm$data:perf_meter.uid"		                /* There is only one UID file for */
  };                                    /* this application. */
static int db_filename_num =
                (sizeof db_filename_vec / sizeof db_filename_vec [0]);

#define PFM_CUSTOMIZE_FILE "DECW$USER_DEFAULTS:DECW$PFM.DAT"
#define SYS_PFM_CUSTOMIZE_FILE "DECW$SYSTEM_DEFAULTS:DECW$PFM.DAT"

static int  main_xpos = 0, main_ypos = 50;
static int update_period = 5;
static long int N10PerSec = -1*(10*1000*1000);
static int TIMER_EFN;
static int  settings_modified_p = FALSE;

static readonly struct {
	char *parameter_name;
	int  parameter_type;
#define PARAM_STRING 0
#define PARAM_INTEGER 1
	char *parameter_value;
	int parameter_size;
	} parameter_specs[] = {
	{"PerfMeter.XPos", PARAM_INTEGER, &main_xpos, sizeof(main_xpos)},
	{"PerfMeter.YPos", PARAM_INTEGER, &main_ypos, sizeof(main_ypos)},
	{"PerfMeter.UpdatePeriod", PARAM_INTEGER, &update_period,
	 sizeof(update_period)},
	{NULL, 0, NULL, 0}
	};

/*
 * Forward declarations
 */

void s_error();
void get_something();
static void set_something();

static void activate_proc();
static void create_proc();
static void cancel_proc();
static void update();

static GC MakeGC();

/* The names and addresses of things that DRM has to bind.  The names do
 * not have to be in alphabetical order.  */

static DRMRegisterArg reglist[] = {
	{"ACTIVATE_PROC", (caddr_t) activate_proc},
	{"CREATE_PROC", (caddr_t) create_proc},
	{"CANCEL_PROC", (caddr_t) cancel_proc}
};

static int reglist_num = (sizeof reglist / sizeof reglist [0]);

static char nodename[16];
static $DESCRIPTOR(NODE,nodename);

/*
 * OS transfer point.  The main routine does all the one-time setup and
 * then calls XtMainLoop.
 */

unsigned int main(argc, argv)
unsigned int argc;                  /* Command line argument count. */
char *argv[];                       /* Pointers to command line args. */
    {
    static long int utime[2];
    register int status, i;
    long int timeup_p;
    XEvent event;
    static char header[128];
    static unsigned short int nlen;
    static struct {
	unsigned short blen, icode;
	char *baddr;
	unsigned short int *rlen;
	} syi_itmlst[] = {
	{sizeof(nodename)-2, SYI$_SCSNODE, nodename, &nlen},
	{0, 0, NULL, NULL} };


    sys$getsyiw(0,0,0,syi_itmlst,0,0,0);

    for (i = nlen; i >= 0 && nodename[i] == ' '; i--) nodename[i] = '\0';
    sprintf(header,"System Performance Meters (V2.0) on %s",nodename);
    NODE.dsc$w_length = strlen(nodename);


    read_sys_customization_file();
    read_customization_file();

    DwtInitializeDRM();                 /* Initialize DRM before initializing
                                        /* the X Toolkit. */

    /* If we had user-defined widgets, we would register them with DRM here. */

    /* Initialize the X Toolkit. We get back a top level shell widget. */

    toplevel_widget = XtInitialize(header,      /* Main window banner text. */
				   "PerfMeter", /* Root class name. */
				   NULL,        /* No option list. */
				   0,           /* Number of options. */
				   &argc,       /* Address of argc */
				   argv);       /* argv */


    /* Open the UID files (the output of the UIL compiler) in the hierarchy*/

    if (DwtOpenHierarchy(db_filename_num, /* Number of files. */
		         db_filename_vec, /* Array of file names.  */
			 NULL,            /* Default OS extenstion. */
			 &s_DRMHierarchy) /* Pointer to returned DRM ID */
				         !=DRMSuccess)
        s_error("can't open hierarchy");

    init_application();

    /* Register the items DRM needs to bind for us. */

    DwtRegisterDRMNames(reglist, reglist_num);

    /* Go get the main part of the application. */
    if (DwtFetchWidget(s_DRMHierarchy, "PERF_METER_MAIN", toplevel_widget,
      &main_window_widget, &dummy_class) != DRMSuccess)
        s_error("can't fetch main window");

    /* Manage the main part and realize everything.  The interface comes up
     * on the display now. */

    XtManageChild(main_window_widget);
    set_something(toplevel_widget,DwtNx,main_xpos);
    set_something(toplevel_widget,DwtNy,main_ypos);
    XtRealizeWidget(toplevel_widget);

    XtAddTimeOut(update_period*1000, update, 0);

    XtMainLoop();
    }

/*
 * One-time initialization of application data structures.
 */

static int init_application()
    {
    int k;

    /* Initialize the application data structures. */
    for (k = 0; k < MAX_WIDGETS; k++)
        widget_array[k] = NULL;
    init_cpu_times();
    }

read_customization_file()
    {
    register char *p, *q;
    char *fgets();
    FILE *custfile;
    static char temp[1024];

    custfile = fopen(PFM_CUSTOMIZE_FILE,"r");
    if (custfile != NULL) {
	while ((p = fgets(temp,1024,custfile)) != NULL) {
	    q = strchr(p,'!'); if (q != NULL) *q = '\0';
	    q = strchr(p,'\n'); if (q != NULL) *q = '\0';
	    process_cust_line(p);
	    }
	fclose(custfile);
	}
    }

read_sys_customization_file()
    {
    register char *p, *q;
    char *fgets();
    FILE *custfile;
    static char temp[1024];

    main_xpos = 0; main_ypos =  95;
    update_period = 5;
    custfile = fopen(SYS_PFM_CUSTOMIZE_FILE,"r");
    if (custfile != NULL) {
	while ((p = fgets(temp,1024,custfile)) != NULL) {
	    q = strchr(p,'!'); if (q != NULL) *q = '\0';
	    q = strchr(p,'\n'); if (q != NULL) *q = '\0';
	    process_cust_line(p);
	    }
	fclose(custfile);
	}
    }

static process_cust_line(p)
register char *p;
    {
    register int i;
    register char *eq;
    char *compare_param();

    while (*p != '\0' && isspace(*p)) p++;
    if (*p == '\0') return;
    for (i = 0; parameter_specs[i].parameter_name != NULL; i++) {
	eq = compare_param(p,parameter_specs[i].parameter_name);
	if (eq != NULL) {
	    switch (parameter_specs[i].parameter_type) {
		case PARAM_STRING:
		    strncpy(parameter_specs[i].parameter_value,eq+1,parameter_specs[i].parameter_size-1);
		    parameter_specs[i].parameter_value[parameter_specs[i].parameter_size-1] = '\0';
		    break;
		case PARAM_INTEGER:
		    *((int *) parameter_specs[i].parameter_value) = atoi(eq+1);
		    break;
	    	}
	    break;
	    }
	}
    }

static char *compare_param(test,pname)
register char *test, *pname;
    {
    register char *p, *q;

    for (p = test, q = pname; *q != '\0'; q++) {
        if (*p == *q) p++;
	else if (*p == '\0') return(NULL);
	else if (*p == '=') return(NULL);
	else if (*p == '*') {
	    while (*(p+1) == '*') p++;
	    if (*(q+1) == '\0') p++;
	    else if (*(q+1) == *(p+1)) p++;
	    }
	else if (*p == '%') p++;
	else return(NULL);
	}
    while (*p != '\0' && isspace(*p)) p++;
    if (*p == '=') return(p);
    else return(NULL);
    }

/***************************************************************************
 *
 * These are some little utilities used by the callback routines.
 */


/*
 * All errors are fatal.
 */

static void s_error(problem_string)
char *problem_string;
    {
    printf("%s\n", problem_string);
    exit(0);
    }

/*
 * Simplified SET VALUE routine to use only when changing a single attribute.
 * If we need to change more than one, all new values should be put
 * into one arglist and we should make one XtSetValues call (which is MUCH
 * more efficient).
 */

static void set_something(w, resource, value)
Widget w;
char *resource, *value;
    {
    Arg al[1];

    XtSetArg(al[0], resource, value);
    XtSetValues(w, al, 1);
    }


/*
 * Simplified GET VALUE routine to use only when retrieving a single attribute.
 * If we need to retrieve more than one, all values should be put
 * into one arglist and we should make one XtGetValues call (which is MUCH
 * more efficient).
 */

static void get_something(w, resource, value)
Widget w;
char *resource, *value;
    {
    Arg al[1];

    XtSetArg(al[0], resource, value);
    XtGetValues(w, al, 1);
    }


int get_widget_if_needed (index, name)
int index;
char *name;
    {
    if (widget_array[index] == NULL) {
	if (DwtFetchWidget(s_DRMHierarchy,name,toplevel_widget,
			   &widget_array[index],&dummy_class) != DRMSuccess)
	    return(FALSE);
	}
    return(TRUE);
    }

/*
 * All widgets that are created call back to this procedure. We just log the
 * ID in the global array.
 */

static void create_proc(w, tag, reason)
Widget w;
int *tag;
unsigned long *reason;
    {
    int widget_num = *tag;

    widget_array[widget_num] = w;

    switch (widget_num) {
	case PFM_CPU_GRAPH:
	    cpu_graph_disp = XtDisplay(w);
	    break;
	case PFM_STATE_GRAPH:
	    state_graph_disp = XtDisplay(w);
	    break;
	case PFM_DEVICE_0_IOCOUNT_NAME:
	    update_disk_name_label(w, 0);
	    break;
	case PFM_DEVICE_1_IOCOUNT_NAME:
	    update_disk_name_label(w, 1);
	    break;
	case PFM_DEVICE_2_IOCOUNT_NAME:
	    update_disk_name_label(w, 2);
	    break;
	case PFM_DEVICE_3_IOCOUNT_NAME:
	    update_disk_name_label(w, 3);
	    break;
        default:
            break;
	}

    }

/*
 * All push buttons in this application call back to this routine.  We
 * use the tag to tell us what widget it is, then react accordingly.
 */

static void activate_proc(w, tag, reason)
Widget w;
int *tag;
unsigned long *reason;
    {
    int widget_num = *tag;              /* Convert tag to widget number. */

    switch (widget_num) {
	case PFM_EXIT_BUTTON: exit();
	default: break;
	}
    }

static void cancel_proc(w, tag, reason)
Widget w;
int *tag;
unsigned long *reason;
    {
    int widget_num = *tag;              /* Convert tag to widget number. */

    switch (widget_num) {
	default: break;
	}
    }

static int CPUPercent[101] = {
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0};

static int Device_0_IO = -1,
	   Device_1_IO = -1,
	   Device_2_IO = -1,
	   Device_3_IO = -1;

static int PageFaults = -1,
	   PageIO = -1,
	   FreePages = -1,
	   ModPages = -1;


#define CompTimeX  2
#define NullTimeX 12
#define UserTimeX 22
#define SupeTimeX 32
#define ExecTimeX 42
#define KernTimeX 52
#define InteTimeX 62

#define StateBarWidth 7

static void update()
    {
    static int NewNullTime,
	   NewUserTime,
	   NewSupeTime,
	   NewExecTime,
	   NewKernTime,
	   NewInteTime,
	   NewCompTime,
	   NewDevice_0_IO, DeltaDevice_0_IO, LastDeltaDevice_0_IO = -1,
	   NewDevice_1_IO, DeltaDevice_1_IO, LastDeltaDevice_1_IO = -1,
	   NewDevice_2_IO, DeltaDevice_2_IO, LastDeltaDevice_2_IO = -1,
	   NewDevice_3_IO, DeltaDevice_3_IO, LastDeltaDevice_3_IO = -1,
	   LastDeltaPageFaults = -1, LastDeltaPageIO = -1,
	   LastDeltaFreePages = -1, LastDeltaModPages = -1;
    register int NewPageFaults, DeltaPageFaults, 
		 NewPageIO, DeltaPageIO, 
		 NewFreePages, DeltaFreePages, 
		 NewModPages, DeltaModPages, NewDispFlag;
    register int total_time, NullTime,
	   UserTime,
	   SupeTime,
	   ExecTime,
	   KernTime,
	   InteTime,
	   CompTime,
	   DeltaIO,ix;
    static char label_buffer[32];
	
    if (cpu_graph_wind == 0) {
	cpu_graph_wind = XtWindow(widget_array[PFM_CPU_GRAPH]);
	cpu_graph_gc1  = MakeGC(cpu_graph_disp, cpu_graph_wind, GXinvert);
	cpu_graph_gc0  = MakeGC(cpu_graph_disp, cpu_graph_wind, GXclear);
	}
    if (state_graph_wind == 0) {
	state_graph_wind = XtWindow(widget_array[PFM_STATE_GRAPH]);
	state_graph_gc1  = MakeGC(state_graph_disp, state_graph_wind, GXinvert);
	state_graph_gc0  = MakeGC(state_graph_disp, state_graph_wind, GXclear);
	}

    get_cpu_times(&NewNullTime,
	   &NewUserTime,
	   &NewSupeTime,
	   &NewExecTime,
	   &NewKernTime,
	   &NewInteTime,
	   &NewCompTime);
    NewPageFaults = current_faults;
    NewPageIO = current_pageio;
    NewFreePages = current_freepages;
    NewModPages = current_modpages;

    total_time = NewNullTime + NewUserTime + NewSupeTime + NewExecTime +
		 NewKernTime + NewInteTime + NewCompTime;
    if (total_time > 0) {
	NullTime = (NewNullTime * 100) / total_time;
	UserTime = (NewUserTime * 100) / total_time;
	SupeTime = (NewSupeTime * 100) / total_time;
	ExecTime = (NewExecTime * 100) / total_time;
	KernTime = (NewKernTime * 100) / total_time;
	InteTime = (NewInteTime * 100) / total_time;
	CompTime = (NewCompTime * 100) / total_time;
	}
    else {
	NullTime = NewNullTime;
	UserTime = NewUserTime;
	SupeTime = NewSupeTime;
	ExecTime = NewExecTime;
	KernTime = NewKernTime;
	InteTime = NewInteTime;
	CompTime = NewCompTime;
	}

    XClearArea(state_graph_disp,state_graph_wind,NullTimeX,0,7,100,0);
    XFillRectangle(state_graph_disp,state_graph_wind,state_graph_gc1,NullTimeX,
								    100-NullTime,
								    7,NullTime);
    XClearArea(state_graph_disp,state_graph_wind,UserTimeX,0,7,100,0);
    XFillRectangle(state_graph_disp,state_graph_wind,state_graph_gc1,UserTimeX,
								    100-UserTime,
								    7,UserTime);
    XClearArea(state_graph_disp,state_graph_wind,SupeTimeX,0,7,100,0);
    XFillRectangle(state_graph_disp,state_graph_wind,state_graph_gc1,SupeTimeX,
								    100-SupeTime,
								    7,SupeTime);
    XClearArea(state_graph_disp,state_graph_wind,ExecTimeX,0,7,100,0);
    XFillRectangle(state_graph_disp,state_graph_wind,state_graph_gc1,ExecTimeX,
								    100-ExecTime,
								    7,ExecTime);
    XClearArea(state_graph_disp,state_graph_wind,KernTimeX,0,7,100,0);
    XFillRectangle(state_graph_disp,state_graph_wind,state_graph_gc1,KernTimeX,
								    100-KernTime,
								    7,KernTime);
    XClearArea(state_graph_disp,state_graph_wind,InteTimeX,0,7,100,0);
    XFillRectangle(state_graph_disp,state_graph_wind,state_graph_gc1,InteTimeX,
								    100-InteTime,
								    7,InteTime);
    XClearArea(state_graph_disp,state_graph_wind,CompTimeX,0,7,100,0);
    XFillRectangle(state_graph_disp,state_graph_wind,state_graph_gc1,CompTimeX,
								    100-CompTime,
								    7,CompTime);

    CPUPercent[100] = 100 - NullTime;
    for (ix = 0; ix < 100; ix++) {
	CPUPercent[ix] = CPUPercent[ix+1];
	XClearArea(cpu_graph_disp,cpu_graph_wind,ix,0,1,100,0);
	XFillRectangle(cpu_graph_disp,cpu_graph_wind,cpu_graph_gc1,ix,100-CPUPercent[ix],
								  1,CPUPercent[ix]);
	}

    if ((widget_array[PFM_DEVICE_0_IOCOUNT_TOTAL] != NULL) &&
	(widget_array[PFM_DEVICE_0_IOCOUNT_DELTA] != NULL))
	{
	get_disk_opcnt(0, &NewDevice_0_IO);
	if ((NewDispFlag = (Device_0_IO < 0))) Device_0_IO = NewDevice_0_IO;
	DeltaDevice_0_IO = NewDevice_0_IO - Device_0_IO;
	if (DeltaDevice_0_IO || NewDispFlag)
		{
		Device_0_IO = NewDevice_0_IO;
		sprintf(label_buffer," %8d",Device_0_IO);
		set_something(widget_array[PFM_DEVICE_0_IOCOUNT_TOTAL],DwtNlabel,
			DwtLatin1String(label_buffer));
		}
	if (DeltaDevice_0_IO != LastDeltaDevice_0_IO)
		{
		if (DeltaDevice_0_IO == 0)
		strcpy(label_buffer,"      ");
		else
		sprintf(label_buffer,"  %4d",DeltaDevice_0_IO);
		set_something(widget_array[PFM_DEVICE_0_IOCOUNT_DELTA],DwtNlabel,
			DwtLatin1String(label_buffer));
		LastDeltaDevice_0_IO = DeltaDevice_0_IO;
		}
	}

    if ((widget_array[PFM_DEVICE_1_IOCOUNT_TOTAL] != NULL) &&
	(widget_array[PFM_DEVICE_1_IOCOUNT_DELTA] != NULL))
	{
	get_disk_opcnt(1, &NewDevice_1_IO);
	if ((NewDispFlag = (Device_1_IO < 0))) Device_1_IO = NewDevice_1_IO;
	DeltaDevice_1_IO = NewDevice_1_IO - Device_1_IO;
	if ((DeltaDevice_1_IO || NewDispFlag) && Device_1_IO >= 0)
	    {
	    Device_1_IO = NewDevice_1_IO;
	    sprintf(label_buffer," %8d",Device_1_IO);
	    set_something(widget_array[PFM_DEVICE_1_IOCOUNT_TOTAL],DwtNlabel,
			DwtLatin1String(label_buffer));
	    }
	if (DeltaDevice_1_IO != LastDeltaDevice_1_IO)
	    {
	    if (DeltaDevice_1_IO == 0)
		strcpy(label_buffer,"      ");
	    else
		sprintf(label_buffer,"  %4d",DeltaDevice_1_IO);
	    set_something(widget_array[PFM_DEVICE_1_IOCOUNT_DELTA],DwtNlabel,
			DwtLatin1String(label_buffer));
	    LastDeltaDevice_1_IO = DeltaDevice_1_IO;
	    }
	}

    if ((widget_array[PFM_DEVICE_2_IOCOUNT_TOTAL] != NULL) &&
	(widget_array[PFM_DEVICE_2_IOCOUNT_DELTA] != NULL))
	{
	get_disk_opcnt(2, &NewDevice_2_IO);
	if ((NewDispFlag = (Device_2_IO < 0))) Device_2_IO = NewDevice_2_IO;
	DeltaDevice_2_IO = NewDevice_2_IO - Device_2_IO;
	if ((DeltaDevice_2_IO || NewDispFlag) && Device_2_IO >= 0)
	    {
	    Device_2_IO = NewDevice_2_IO;
	    sprintf(label_buffer," %8d",Device_2_IO);
	    set_something(widget_array[PFM_DEVICE_2_IOCOUNT_TOTAL],DwtNlabel,
			DwtLatin1String(label_buffer));
	    }
	if (DeltaDevice_2_IO != LastDeltaDevice_2_IO)
	    {
	    if (DeltaDevice_2_IO == 0)
		strcpy(label_buffer,"      ");
	    else
		sprintf(label_buffer,"  %4d",DeltaDevice_2_IO);
	    set_something(widget_array[PFM_DEVICE_2_IOCOUNT_DELTA],DwtNlabel,
			DwtLatin1String(label_buffer));
	    LastDeltaDevice_2_IO = DeltaDevice_2_IO;
	    }
	}

    if ((widget_array[PFM_DEVICE_3_IOCOUNT_TOTAL] != NULL) &&
	(widget_array[PFM_DEVICE_3_IOCOUNT_DELTA] != NULL))
	{
	get_disk_opcnt(3, &NewDevice_3_IO);
	if ((NewDispFlag = (Device_3_IO < 0))) Device_3_IO = NewDevice_3_IO;
	DeltaDevice_3_IO = NewDevice_3_IO - Device_3_IO;
	if ((DeltaDevice_3_IO || NewDispFlag) && Device_3_IO >= 0)
	    {
	    Device_3_IO = NewDevice_3_IO;
	    sprintf(label_buffer," %8d",Device_3_IO);
	    set_something(widget_array[PFM_DEVICE_3_IOCOUNT_TOTAL],DwtNlabel,
			DwtLatin1String(label_buffer));
	    }
	if (DeltaDevice_3_IO != LastDeltaDevice_3_IO)
	    {
	    if (DeltaDevice_3_IO == 0)
		strcpy(label_buffer,"      ");
	    else
		sprintf(label_buffer,"  %4d",DeltaDevice_3_IO);
	    set_something(widget_array[PFM_DEVICE_3_IOCOUNT_DELTA],DwtNlabel,
			DwtLatin1String(label_buffer));
	    LastDeltaDevice_3_IO = DeltaDevice_3_IO;
	    }
	}

    if ((NewDispFlag = (PageFaults < 0))) PageFaults = NewPageFaults;
    DeltaPageFaults = NewPageFaults - PageFaults;
    if (DeltaPageFaults || NewDispFlag)
	{
	PageFaults = NewPageFaults;
	sprintf(label_buffer," %8d",PageFaults);
	set_something(widget_array[PFM_PAGEFAULTS_TOTAL],DwtNlabel,
		    DwtLatin1String(label_buffer));
	}
    if (DeltaPageFaults != LastDeltaPageFaults)
	{
	if (DeltaPageFaults == 0)
	    strcpy(label_buffer,"      ");
	else
	    sprintf(label_buffer,"  %4d",DeltaPageFaults);
	set_something(widget_array[PFM_PAGEFAULTS_DELTA],DwtNlabel,
		    DwtLatin1String(label_buffer));
	LastDeltaPageFaults = DeltaPageFaults;
	}

    if ((NewDispFlag = (PageIO < 0))) PageIO = NewPageIO;
    DeltaPageIO = NewPageIO - PageIO;
    if (DeltaPageIO || NewDispFlag)
	{
	PageIO = NewPageIO;
	sprintf(label_buffer," %8d",PageIO);
	set_something(widget_array[PFM_PAGEIO_TOTAL],DwtNlabel,
		    DwtLatin1String(label_buffer));
	}
    if (DeltaPageIO != LastDeltaPageIO)
	{
	if (DeltaPageIO == 0)
	    strcpy(label_buffer,"      ");
	else
	    sprintf(label_buffer,"  %4d",DeltaPageIO);
	set_something(widget_array[PFM_PAGEIO_DELTA],DwtNlabel,
		    DwtLatin1String(label_buffer));
	LastDeltaPageIO = DeltaPageIO;
	}

    if ((NewDispFlag = (FreePages < 0))) FreePages = NewFreePages;
    DeltaFreePages = NewFreePages - FreePages;
    if (DeltaFreePages || NewDispFlag)
	{
	FreePages = NewFreePages;
	sprintf(label_buffer," %8d",FreePages);
	set_something(widget_array[PFM_FREEPAGES_TOTAL],DwtNlabel,
		    DwtLatin1String(label_buffer));
	}
    if (DeltaFreePages != LastDeltaFreePages)
	{
	if (DeltaFreePages == 0)
	    strcpy(label_buffer,"      ");
	else
	    sprintf(label_buffer,"  %4d",DeltaFreePages);
	set_something(widget_array[PFM_FREEPAGES_DELTA],DwtNlabel,
		    DwtLatin1String(label_buffer));
	LastDeltaFreePages = DeltaFreePages;
	}

    if ((NewDispFlag = (ModPages < 0))) ModPages = NewModPages;
    DeltaModPages = NewModPages - ModPages;
    if (DeltaModPages || NewDispFlag)
	{
	ModPages = NewModPages;
	sprintf(label_buffer," %8d",ModPages);
	set_something(widget_array[PFM_MODPAGES_TOTAL],DwtNlabel,
		    DwtLatin1String(label_buffer));
	}
    if (DeltaModPages != LastDeltaModPages)
	{
	if (DeltaModPages == 0)
	    strcpy(label_buffer,"      ");
	else
	    sprintf(label_buffer,"  %4d",DeltaModPages);
	set_something(widget_array[PFM_MODPAGES_DELTA],DwtNlabel,
		      DwtLatin1String(label_buffer));
	LastDeltaModPages = DeltaModPages;
	}

    XtAddTimeOut(update_period*1000, update, 0);
    }

static GC MakeGC(disp,wind,fun)
Display *disp;
Window wind;
int fun;
{
    static XGCValues gcvalues = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    gcvalues.function = fun;
    return(XCreateGC(disp,wind,GCFunction,&gcvalues));
    }

static long int last_user, last_super, last_exec, last_kernel,
		last_interrupt, last_compat, last_spin, last_idle;

#define K_TIME 0
#define E_TIME 1
#define S_TIME 2
#define U_TIME 3
#define I_TIME 4
#define C_TIME 5
#define SPIN_TIME 6
#define IDLE_TIME 7

static init_cpu_times()
    {
    register int i, j, k;
    long int *PMS$GL_KERNEL;

    last_user = last_super = last_exec = last_kernel = last_interrupt =
	last_compat = last_spin = last_idle = 0;

    for (k = SMP$GL_ACTIVE_CPUS, j = 0; k != 0; j++, k = k >> 1) {
	if ((k & 01) == 0) continue;
	PMS$GL_KERNEL = (SMP$GL_CPU_DATA[j])->cpu$l_kernel;
	last_kernel += PMS$GL_KERNEL[K_TIME];
	last_exec += PMS$GL_KERNEL[E_TIME];
	last_super += PMS$GL_KERNEL[S_TIME];
	last_user += PMS$GL_KERNEL[U_TIME];
	last_interrupt += PMS$GL_KERNEL[I_TIME];
	last_compat += PMS$GL_KERNEL[C_TIME];
	last_spin += PMS$GL_KERNEL[SPIN_TIME];
	last_idle += PMS$GL_KERNEL[IDLE_TIME];
	}
    last_interrupt -= last_idle;
    }

static get_cpu_times(null,user,super,exec,kernel,interrupt,compat)
int *null,*user,*super,*exec,*kernel,*interrupt,*compat;
    {
    register long int new_null, new_user, new_super, new_exec,
		    new_kernel, new_interrupt, new_compat,
		    new_idle, new_spin;
    register int i, j, k;
    long int *PMS$GL_KERNEL;
    register long int new_abstime, totaltime;

    new_user = new_super = new_exec = new_kernel = new_interrupt =
	new_compat = new_spin = new_idle = 0;

    for (k = SMP$GL_ACTIVE_CPUS, j = 0; k != 0; j++, k = k >> 1) {
	if ((k & 01) == 0) continue;
	PMS$GL_KERNEL = (SMP$GL_CPU_DATA[j])->cpu$l_kernel;
	new_kernel += PMS$GL_KERNEL[K_TIME];
	new_exec += PMS$GL_KERNEL[E_TIME];
	new_super += PMS$GL_KERNEL[S_TIME];
	new_user += PMS$GL_KERNEL[U_TIME];
	new_interrupt += PMS$GL_KERNEL[I_TIME];
	new_compat += PMS$GL_KERNEL[C_TIME];
	new_spin += PMS$GL_KERNEL[SPIN_TIME];
	new_idle += PMS$GL_KERNEL[IDLE_TIME];
	}
    new_interrupt -= new_idle;

    *user = new_user - last_user; last_user = new_user;
    *super = new_super - last_super; last_super = new_super;
    *exec = new_exec - last_exec; last_exec = new_exec;
    *kernel = new_kernel - last_kernel; last_kernel = new_kernel;
    *interrupt = new_interrupt - last_interrupt; last_interrupt = new_interrupt;
    *compat = new_compat - last_compat; last_compat = new_compat;

    *null = new_idle - last_idle; last_idle = new_idle;
    }

static get_disk_opcnt(devnum, count)
register int devnum;
register int *count;
    {
    static struct {
	unsigned short int blen, icode;
	int *buff;
	unsigned short int *rlen;
	} dvi_itmlst[] = {
	{sizeof(int), DVI$_OPCNT, NULL, NULL},
	{0, 0, NULL, NULL}
	};
    static struct {
	unsigned short int istat, bytes;
	long int fill;
	} iosb;
    char devname[64];
    $DESCRIPTOR(dev,devname);
    int status;

    sprintf(devname, "PFM$DEVICE_%d_NAME", devnum);
    dvi_itmlst[0].buff = count;
    *count = -1;
    dev.dsc$w_length = strlen(devname);
    status = sys$getdviw(0,0,&dev,dvi_itmlst,&iosb,0,0,0);
    if ((status == SS$_IVDEVNAM) && (devnum == 0))
	{
	$DESCRIPTOR(sysdev, "SYS$SYSDEVICE:");

	sys$getdviw(0,0,&sysdev,dvi_itmlst,&iosb,0,0,0);
	}
    }


static update_disk_name_label(w, devnum)
Widget w;
int devnum;
    {
    char devname[64];
    char *title;

    sprintf(devname, "PFM$DEVICE_%d_TITLE", devnum);
    title = getenv(devname);
    if (NULL == title)
	{
	sprintf(devname, "PFM$DEVICE_%d_NAME", devnum);
	title = getenv(devname);
	if ((NULL == title) && (0 == devnum))
	    title = "SYS$SYSDEVICE:";
	}
    if (title != NULL) set_something(w,DwtNlabel,DwtLatin1String(title));
    }
